#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/tads3/resload.cpp,v 1.2 1999/05/17 02:52:29 MJRoberts Exp $";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  resload.cpp - resource loader
Function
  
Notes
  
Modified
  10/17/98 MJRoberts  - Creation
*/

#include <stddef.h>
#include <string.h>

#include "t3std.h"
#include "resload.h"
#include "vmimage.h"

/* ------------------------------------------------------------------------ */
/*
 *   Resource loader interface implementation 
 */
class CVmImageLoaderMres_resload: public CVmImageLoaderMres
{
public:
    CVmImageLoaderMres_resload(const char *respath)
    {
        /* remember the path of the resource we're trying to find */
        respath_ = respath;
        respath_len_ = strlen(respath);

        /* we haven't found it yet */
        found_ = FALSE;
    }

    /* add a resource */
    void add_resource(uint32 seek_ofs, uint32 siz,
                      const char *res_name, size_t res_name_len)
    {
        /* 
         *   if we've already found a match, there's no need to consider
         *   anything else 
         */
        if (found_)
            return;

        /* check to see if this is the one we're looking for */
        if (res_name_len == respath_len_
            && memicmp(respath_, res_name, res_name_len) == 0)
        {
            /* we found it */
            found_ = TRUE;

            /* remember the seek location */
            res_seek_ = seek_ofs;
            res_size_ = siz;
        }
    }

    /* did we find the resource? */
    int found_resource() const { return found_; }

    /* get the seek location and size of the resource we found */
    uint32 get_resource_seek() const { return res_seek_; }
    uint32 get_resource_size() const { return res_size_; }

private:
    /* name of the resource we're looking for */
    const char *respath_;
    size_t respath_len_;

    /* flag: we found the resource we're looking for */
    int found_;

    /* seek location and size of the resource we found */
    uint32 res_seek_;
    uint32 res_size_;
};

/* ------------------------------------------------------------------------ */
/*
 *   create resource loader 
 */
CResLoader::CResLoader()
{
    /* we have no root directory */
    root_dir_ = 0;

    /* no executable path yet */
    exe_filename_ = 0;
}

/*
 *   create a resource loader given a root directory 
 */
CResLoader::CResLoader(const char *root_dir)
{
    /* remember the root directory for file searches */
    root_dir_ = lib_copy_str(root_dir);

    /* no executable path yet */
    exe_filename_ = 0;
}

/*
 *   delete resource loader 
 */
CResLoader::~CResLoader()
{
    /* free our root directory string and executable path string */
    lib_free_str(root_dir_);
    lib_free_str(exe_filename_);
}

/*
 *   Open a resource file given the resource path 
 */
osfildef *CResLoader::open_res_file(const char *respath,
                                    const char *restype)
{
    char filepath[OSFNMAX];
    osfildef *fp;
    
    /* 
     *   Look for the resource as an external file.  If we have a root
     *   directory, look for the file under that directory; otherwise,
     *   look for it in the current directory.  
     */
    if (root_dir_ != 0)
    {
        char fname[OSFNMAX];
        
        /* get the resource name as a file path */
        os_cvt_url_dir(fname, sizeof(fname), respath, FALSE);

        /* build a full path from the root directory and the resource path */
        os_build_full_path(filepath, sizeof(filepath), root_dir_, fname);
    }
    else
    {
        /* get the resource name as a file path */
        os_cvt_url_dir(filepath, sizeof(filepath), respath, FALSE);
    }

    /* try opening the file */
    fp = osfoprb(filepath, OSFTBIN);

    /* 
     *   if that failed, and we had a root directory to look in, try again
     *   in the current directory 
     */
    if (fp == 0 && root_dir_ != 0)
    {
        /* get the resource name as a simple file path */
        os_cvt_url_dir(filepath, sizeof(filepath), respath, FALSE);

        /* try opening it */
        fp = osfoprb(filepath, OSFTBIN);
    }

    /*
     *   If we couldn't find an external resource, try looking in the
     *   executable file. 
     */
    if (fp == 0 && exe_filename_ != 0 && restype != 0)
    {
        osfildef *exe_fp;
        
        /* find the executable file's resources */
        exe_fp = os_exeseek(exe_filename_, restype);

        /* if we found something, try loading from that file */
        if (exe_fp != 0)
        {
            CVmImageLoaderMres_resload res_ifc(respath);

            /* try loading the resources */
            CVmImageLoader::
                load_resources_from_fp(exe_fp, exe_filename_, &res_ifc);

            /* check to see if we found it */
            if (res_ifc.found_resource())
            {
                /* we got it - use the executable file as the resource file */
                fp = exe_fp;

                /* seek to the location of the resource */
                osfseek(fp, res_ifc.get_resource_seek(), OSFSK_SET);
            }
            else
            {
                /* didn't find it - forget the executable file */
                osfcls(exe_fp);
            }
        }
    }

    /* return the result */
    return fp;
}

