/* $Header: d:/cvsroot/tads/tads3/resload.h,v 1.2 1999/05/17 02:52:29 MJRoberts Exp $ */

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  resload.h - resource loader
Function
  
Notes
  
Modified
  10/17/98 MJRoberts  - Creation
*/

#ifndef RESLOAD_H
#define RESLOAD_H

#include "t3std.h"

/* ------------------------------------------------------------------------ */
/*
 *   Resource loader.  
 */

class CResLoader
{
public:
    /* create the loader */
    CResLoader();

    /* 
     *   create the loader with a root directory - we'll look for external
     *   files under this directory 
     */
    CResLoader(const char *root_dir);

    /* set the executable file name */
    void set_exe_filename(const char *exe_filename)
    {
        /* discard the old executable filename, if there is one */
        lib_free_str(exe_filename_);

        /* remember the new path */
        exe_filename_ = lib_copy_str(exe_filename);
    }

    /* delete */
    ~CResLoader();

    /* 
     *   Load a resource given a resource path.  If 'restype' is non-null,
     *   we'll look in the executable file for a resource of the given
     *   type (this should be a four-byte identifier, as used with
     *   MAKETRX) if we can't find an external file for the given path.
     *   This allows resources of different types to be bundled into the
     *   executable file.  Note that resources bundled into the executable
     *   must use the standard T3 resource-image file format.  
     */
    osfildef *open_res_file(const char *respath, const char *restype);

protected:
    /* root directory for external file searches */
    char *root_dir_;

    /* path to executable file */
    char *exe_filename_;
};

#endif /* RESLOAD_H */

