#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/tads3/VMBIFTAD.CPP,v 1.3 1999/07/11 00:46:58 MJRoberts Exp $";
#endif

/* Copyright (c) 1999 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmbiftad.cpp - TADS built-in function set for T3 VM
Function
  
Notes
  
Modified
  04/05/99 MJRoberts  - Creation
*/

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "t3std.h"
#include "os.h"
#include "utf8.h"
#include "vmuni.h"
#include "vmbiftad.h"
#include "vmstack.h"
#include "vmerr.h"
#include "vmerrnum.h"
#include "vmglob.h"
#include "vmpool.h"
#include "vmobj.h"
#include "vmstr.h"
#include "vmlst.h"
#include "vmrun.h"
#include "vmregex.h"
#include "vmundo.h"
#include "vmfile.h"
#include "vmsave.h"
#include "vmbignum.h"



/* ------------------------------------------------------------------------ */
/*
 *   forward statics 
 */

#ifdef VMBIFTADS_RNG_ISAAC
static void isaac_init(isaacctx *ctx, int flag);
#endif /* VMBIFTADS_RNG_ISAAC */


/* ------------------------------------------------------------------------ */
/*
 *   Initialize the TADS intrinsics global state 
 */
CVmBifTADSGlobals::CVmBifTADSGlobals()
{
    int i;
    
    /* allocate our regular exprssion parser */
    regex = new CRegexParser();

    /* clear all of the file slot handles */
    for (i = 0 ; i < VMBIFTADS_FILES_MAX ; ++i)
    {
        /* no file yet */
        fp[i].fp = 0;

        /* no character map yet */
        fp[i].charmap_read = 0;
        fp[i].charmap_write = 0;
    }

#ifdef VMBIFTADS_RNG_LCG
    /* 
     *   Set the random number seed to a fixed starting value (this value
     *   is arbitrary; we chose it by throwing dice).  If the program
     *   wants another sequence, it can manually change this by calling
     *   the randomize() intrinsic in our function set, which seeds the
     *   generator with an OS-dependent starting value (usually based on
     *   the system's real-time clock, to ensure that each run will use a
     *   different starting value).  
     */
    rand_seed = 024136543305;
#endif

#ifdef VMBIFTADS_RNG_ISAAC
    /* create the ISAAC context structure */
    isaac_ctx = (struct isaacctx *)t3malloc(sizeof(struct isaacctx));

    /* initialize with a fixed seed vector */
    isaac_init(isaac_ctx, FALSE);
#endif
}

/*
 *   delete the TADS intrinsics global state 
 */
CVmBifTADSGlobals::~CVmBifTADSGlobals()
{
    int i;
    
    /* delete our regular expression parser */
    delete regex;

    /* close any open files */
    for (i = 0 ; i < VMBIFTADS_FILES_MAX ; ++i)
        fp[i].close();

#ifdef VMBIFTADS_RNG_ISAAC
    /* delete the ISAAC context */
    t3free(isaac_ctx);
#endif
}

/* ------------------------------------------------------------------------ */
/*
 *   datatype - get the datatype of a given value
 */
void CVmBifTADS::datatype(VMG_ uint argc)
{
    vm_val_t val;
    vm_val_t retval;

    /* check arguments */
    check_argc(vmg_ argc, 1);

    /* pop the value */
    G_stk->pop(&val);

    /* return the appropriate value for this type */
    retval.set_datatype(vmg_ &val);
    retval_int(vmg_ retval.val.intval);
}

/* ------------------------------------------------------------------------ */
/*
 *   getarg - get the given argument to the current procedure 
 */
void CVmBifTADS::getarg(VMG_ uint argc)
{
    int idx;

    /* check arguments */
    check_argc(vmg_ argc, 1);

    /* get the argument index value */
    idx = pop_int_val(vmg0_);

    /* if the argument index is out of range, throw an error */
    if (idx < 1 || idx > G_interpreter->get_cur_argc(vmg0_))
        err_throw(VMERR_BAD_VAL_BIF);

    /* push the parameter value */
    *G_interpreter->get_r0() = *G_interpreter->get_param(vmg_ idx - 1);
}

/* ------------------------------------------------------------------------ */
/*
 *   firstobj - get the first object instance
 */
void CVmBifTADS::firstobj(VMG_ uint argc)
{
    /* check arguments */
    check_argc_range(vmg_ argc, 0, 1);

    /* enumerate objects starting with object 1 */
    enum_objects(vmg_ argc, (vm_obj_id_t)1);
}

/*
 *   nextobj - get the next object instance after a given object
 */
void CVmBifTADS::nextobj(VMG_ uint argc)
{
    vm_val_t val;
    vm_obj_id_t prv_obj;

    /* check arguments */
    check_argc_range(vmg_ argc, 1, 2);

    /* get the previous object */
    G_interpreter->pop_obj(vmg_ &val);
    prv_obj = val.val.obj;

    /* 
     *   Enumerate objects starting with object 1.  Reduce the argument
     *   count by one, since we've removed the preceding object. 
     */
    enum_objects(vmg_ argc - 1, prv_obj + 1);
}

/*
 *   Common handler for firstobj/nextobj object iteration
 */
void CVmBifTADS::enum_objects(VMG_ uint argc, vm_obj_id_t start_obj)
{
    vm_val_t val;
    vm_obj_id_t sc;
    vm_obj_id_t obj;

    /* if there's an object whose instances we are to enumerate, get it */
    if (argc == 1)
    {
        /* pop the object */
        G_interpreter->pop_obj(vmg_ &val);
        sc = val.val.obj;
    }

    /* presume we won't find anything */
    retval_nil(vmg0_);

    /* 
     *   starting at object ID 1, scan objects until we find one that's
     *   valid and matches our superclass, if one was provided 
     */
    for (obj = start_obj ; obj < G_obj_table->get_max_used_obj_id() ; ++obj)
    {
        /* if it's valid, and it's not a class object, consider it further */
        if (G_obj_table->is_obj_id_valid(obj)
            && !vm_objp(vmg_ obj)->is_class_object(vmg_ obj))
        {
            /* 
             *   if a superclass was specified, and it matches, we have a
             *   winner 
             */
            if (argc == 1)
            {
                /* if the object matches, return it */
                if (vm_objp(vmg_ obj)->is_instance_of(vmg_ sc))
                {
                    retval_obj(vmg_ obj);
                    break;
                }
            }
            else
            {
                /* 
                 *   we're enumerating all objects - but skip List and
                 *   String metaclass objects 
                 */
                if (vm_objp(vmg_ obj)->get_as_list() == 0
                    && vm_objp(vmg_ obj)->get_as_string(vmg0_) == 0)
                {
                    retval_obj(vmg_ obj);
                    break;
                }
            }
        }
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Random number generators.  Define one of the following configuration
 *   variables to select a random number generation algorithm:
 *   
 *   VMBIFTADS_RNG_LCG - linear congruential generator
 *.  VMBIFTADS_RNG_ISAAC - ISAAC (cryptographic hash generator) 
 */

/* ------------------------------------------------------------------------ */
/*
 *   Linear Congruential Random-Number Generator.  This generator uses an
 *   algorithm from Knuth, The Art of Computer Programming, Volume 2, p.
 *   170, with parameters chosen from the same book for their good
 *   statistical properties and efficiency on 32-bit hardware.  
 */
#ifdef VMBIFTADS_RNG_LCG
/*
 *   randomize - seed the random-number generator 
 */
void CVmBifTADS::randomize(VMG_ uint argc)
{
    /* check arguments */
    check_argc(vmg_ argc, 0);

    /* seed the generator */
    os_rand(&G_bif_tads_globals->rand_seed);
}

/*
 *   generate the next random number - linear congruential generator 
 */
static ulong rng_next(VMG0_)
{
    const ulong a = 1664525L;
    const ulong c = 1;

    /* 
     *   Generate the next random value using the linear congruential
     *   method described in Knuth, The Art of Computer Programming,
     *   volume 2, p170.
     *   
     *   Use 2^32 as m, hence (n mod m) == (n & 0xFFFFFFFF).  This is
     *   efficient and is well-suited to 32-bit machines, works fine on
     *   larger machines, and will even work on 16-bit machines as long as
     *   the compiler can provide us with 32-bit arithmetic (which we
     *   assume extensively elsewhere anyway).
     *   
     *   We use a = 1664525, a multiplier which has very good results with
     *   the Spectral Test (see Knuth p102) with our choice of m.
     *   
     *   Use c = 1, since this trivially satisfies Knuth's requirements
     *   about common factors.
     *   
     *   Note that the result of the multiplication might overflow a
     *   32-bit ulong for values of rand_seed that are not small.  This
     *   doesn't matter, since if it does, the machine will naturally
     *   truncate high-order bits to yield the result mod 2^32.  So, on a
     *   32-bit machine, the (&0xFFFFFFFF) part is superfluous, but it's
     *   harmless and is needed for machines with a larger word size.  
     */
    G_bif_tads_globals->rand_seed =
        (long)(((a * (ulong)G_bif_tads_globals->rand_seed) + 1) & 0xFFFFFFFF);
    return (ulong)G_bif_tads_globals->rand_seed;
}
#endif /* VMBIFTADS_RNG_LCG */

/* ------------------------------------------------------------------------ */
/*
 *   ISAAC random number generator. 
 */

#ifdef VMBIFTADS_RNG_ISAAC

/* service macros for ISAAC random number generator */
#define isaac_ind(mm,x)  ((mm)[(x>>2)&(ISAAC_RANDSIZ-1)])
#define isaac_step(mix,a,b,mm,m,m2,r,x) \
{ \
    x = *m;  \
    a = ((a^(mix)) + *(m2++)) & 0xffffffff; \
    *(m++) = y = (isaac_ind(mm,x) + a + b) & 0xffffffff; \
    *(r++) = b = (isaac_ind(mm,y>>ISAAC_RANDSIZL) + x) & 0xffffffff; \
}
#define isaac_rand(r) \
    ((r)->cnt-- == 0 ? \
    (isaac_gen_group(r), (r)->cnt=ISAAC_RANDSIZ-1, (r)->rsl[(r)->cnt]) : \
    (r)->rsl[(r)->cnt])

#define isaac_mix(a,b,c,d,e,f,g,h) \
{ \
    a^=b<<11; d+=a; b+=c; \
    b^=c>>2;  e+=b; c+=d; \
    c^=d<<8;  f+=c; d+=e; \
    d^=e>>16; g+=d; e+=f; \
    e^=f<<10; h+=e; f+=g; \
    f^=g>>4;  a+=f; g+=h; \
    g^=h<<8;  b+=g; h+=a; \
    h^=a>>9;  c+=h; a+=b; \
}

/* generate the group of numbers */
static void isaac_gen_group(isaacctx *ctx)
{
    ulong a;
    ulong b;
    ulong x;
    ulong y;
    ulong *m;
    ulong *mm;
    ulong *m2;
    ulong *r;
    ulong *mend;
    
    mm = ctx->mem;
    r = ctx->rsl;
    a = ctx->a;
    b = (ctx->b + (++ctx->c)) & 0xffffffff;
    for (m = mm, mend = m2 = m + (ISAAC_RANDSIZ/2) ; m<mend ; )
    {
        isaac_step(a<<13, a, b, mm, m, m2, r, x);
        isaac_step(a>>6,  a, b, mm, m, m2, r, x);
        isaac_step(a<<2,  a, b, mm, m, m2, r, x);
        isaac_step(a>>16, a, b, mm, m, m2, r, x);
    }
    for (m2 = mm; m2<mend; )
    {
        isaac_step(a<<13, a, b, mm, m, m2, r, x);
        isaac_step(a>>6,  a, b, mm, m, m2, r, x);
        isaac_step(a<<2,  a, b, mm, m, m2, r, x);
        isaac_step(a>>16, a, b, mm, m, m2, r, x);
    }
    ctx->b = b;
    ctx->a = a;
}

/* 
 *   Initialize.  If flag is true, then use the contents of ctx->rsl[] to
 *   initialize ctx->mm[]; otherwise, we'll use a fixed starting
 *   configuration.  
 */
static void isaac_init(isaacctx *ctx, int flag)
{
    int i;
    ulong a;
    ulong b;
    ulong c;
    ulong d;
    ulong e;
    ulong f;
    ulong g;
    ulong h;
    ulong *m;
    ulong *r;
    
    ctx->a = ctx->b = ctx->c = 0;
    m = ctx->mem;
    r = ctx->rsl;
    a = b = c = d = e = f = g = h = 0x9e3779b9;         /* the golden ratio */
    
    /* scramble the initial settings */
    for (i = 0 ; i < 4 ; ++i)
    {
        isaac_mix(a, b, c, d, e, f, g, h);
    }

    if (flag) 
    {
        /* initialize using the contents of ctx->rsl[] as the seed */
        for (i = 0 ; i < ISAAC_RANDSIZ ; i += 8)
        {
            a += r[i];   b += r[i+1]; c += r[i+2]; d += r[i+3];
            e += r[i+4]; f += r[i+5]; g += r[i+6]; h += r[i+7];
            isaac_mix(a, b, c, d, e, f, g, h);
            m[i] = a;   m[i+1] = b; m[i+2] = c; m[i+3] = d;
            m[i+4] = e; m[i+5] = f; m[i+6] = g; m[i+7] = h;
        }

        /* do a second pass to make all of the seed affect all of m */
        for (i = 0 ; i < ISAAC_RANDSIZ ; i += 8)
        {
            a += m[i];   b += m[i+1]; c += m[i+2]; d += m[i+3];
            e += m[i+4]; f += m[i+5]; g += m[i+6]; h += m[i+7];
            isaac_mix(a, b, c, d, e, f, g, h);
            m[i] = a;   m[i+1] = b; m[i+2] = c; m[i+3] = d;
            m[i+4] = e; m[i+5] = f; m[i+6] = g; m[i+7] = h;
        }
    }
    else
    {
        /* initialize using fixed initial settings */
        for (i = 0 ; i < ISAAC_RANDSIZ ; i += 8)
        {
            isaac_mix(a, b, c, d, e, f, g, h);
            m[i] = a; m[i+1] = b; m[i+2] = c; m[i+3] = d;
            m[i+4] = e; m[i+5] = f; m[i+6] = g; m[i+7] = h;
        }
    }

    /* fill in the first set of results */
    isaac_gen_group(ctx);

    /* prepare to use the first set of results */    
    ctx->cnt = ISAAC_RANDSIZ;
}

/*
 *   seed the rng 
 */
void CVmBifTADS::randomize(VMG_ uint argc)
{
    int i;
    long seed;
    
    /* check arguments */
    check_argc(vmg_ argc, 0);

    /* seed the generator */
    os_rand(&seed);

    /* 
     *   Fill in rsl[] with the seed. It doesn't do a lot of good to call
     *   os_rand() repeatedly, since this function might simply return the
     *   real-time clock value.  So, use the os_rand() seed value as the
     *   first rsl[] value, then use a simple linear congruential
     *   generator to fill in the rest of rsl[].  
     */
    for (i = 0 ; i < ISAAC_RANDSIZ ; ++i)
    {
        const ulong a = 1664525L;
        const ulong c = 1;

        /* fill in this value from the previous seed value */
        G_bif_tads_globals->isaac_ctx->rsl[i] = (ulong)seed;

        /* generate the next lcg value */
        seed = (long)(((a * (ulong)seed) + 1) & 0xFFFFFFFF);
    }

    /* initialize with this rsl[] array */
    isaac_init(G_bif_tads_globals->isaac_ctx, TRUE);
}


/*
 *   generate the next random number - ISAAC (by Bob Jenkins,
 *   http://ourworld.compuserve.com/homepages/bob_jenkins/isaacafa.htm) 
 */
static ulong rng_next(VMG0_)
{
    /* return the next number */
    return isaac_rand(G_bif_tads_globals->isaac_ctx);
}
#endif /* VMBIFTADS_RNG_ISAAC */

/* ------------------------------------------------------------------------ */
/*
 *   rand - generate a random number, or choose an element randomly from a
 *   list of values or from our list of arguments.
 *   
 *   With one integer argument N, we choose a random number from 0 to N-1.
 *   
 *   With one list argument, we choose a random element of the list.
 *   
 *   With multiple arguments, we choose one argument at random and return
 *   its value.  Note that, because this is an ordinary built-in function,
 *   all of our arguments will be fully evaluated.  
 */
void CVmBifTADS::rand(VMG_ uint argc)
{
    ulong range;
    int choose_an_arg;
    const char *listp;
    ulong rand_val;

    /* presume we're not going to choose from our arguments or from a list */
    choose_an_arg = FALSE;
    listp = 0;

    /* determine the desired range of values based on the arguments */
    if (argc == 0)
    {
        /* 
         *   if no argument is given, produce a random number in our full
         *   range; indicate this by using 0 for our range
         */
        range = 0;
    }
    else if (argc == 1 && G_stk->get(0)->typ == VM_INT)
    {
        /* we're returning a number in the range 0..(arg-1) */
        range = G_stk->get(0)->val.intval;

        /* discard the argument */
        G_stk->discard();
    }
    else if (argc == 1)
    {
        /* pop the list value */
        listp = pop_list_val(vmg0_);

        /* our range is 0..(list_element_count-1) */
        range = vmb_get_len(listp);
    }
    else
    {
        /* 
         *   produce a random number in the range 0..(argc-1) so that we
         *   can select one of our arguments 
         */
        range = argc;

        /* note that we should choose an argument value */
        choose_an_arg = TRUE;
    }

    /* get the next random number */
    rand_val = rng_next(vmg0_);

    /*
     *   Calculate our random value in the range 0..(range-1).  If range
     *   == 0, simply choose a value across our full range.  
     */
    if (range != 0)
    {
        /* 
         *   A range was specified, so choose in our range.  As Knuth
         *   suggests, don't simply take the low-order bits from the
         *   value, since these are the least random part.  Instead, use
         *   the method Knuth describes in TAOCP Vol 2 section 3.4.2. 
         */
        rand_val = (ulong)((((double)rand_val) / 4294967296.0)
                           * (double)range);
    }

    /*
     *   Return the appropriate value, depending on our argument list 
     */
    if (choose_an_arg)
    {
        /* return the selected argument */
        retval(vmg_ G_stk->get((int)rand_val));

        /* discard all of the arguments */
        G_stk->discard(argc);
    }
    else if (listp != 0)
    {
        vm_val_t val;
        
        /* get the selected list element */
        vmb_get_dh(listp + VMB_LEN + (size_t)((rand_val * VMB_DATAHOLDER)),
                   &val);

        /* set the result */
        retval(vmg_ &val);
    }
    else
    {
        /* simply return the random number */
        retval_int(vmg_ (long)rand_val);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Bit-shift generator.  This is from Knuth, The Art of Computer
 *   Programming, volume 2.  This generator is designed to produce random
 *   strings of bits and is not suitable for use as a general-purpose RNG.
 *   
 *   Linear congruential generators are not ideal for generating random
 *   bits; their statistical properties seem better suited for generating
 *   values over a wider range.  This generator is specially designed to
 *   produce random bits, so it could be a useful complement to an LCG RNG.
 *   
 *   This code should not be enabled in its present state; it's retained
 *   in case we want in the future to implement a generator exclusively
 *   for random bits.  The ISAAC generator seems to be a good source of
 *   random bits as well as random numbers, so it seems unlikely that
 *   we'll need a separate random bit generator.  
 */

#ifdef VMBIFTADS_RNG_BITSHIFT
void CVmBifTADS::randbit(VMG_ uint argc)
{
    int top_bit;
    
    /* check arguments */
    check_argc(vmg_ argc, 0);

    top_bit = (G_bif_tads_globals->rand_seed & 0x8000000);
    G_bif_tads_globals->rand_seed <<= 1;
    if (top_bit)
        G_bif_tads_globals->rand_seed ^= 035604231625;

    retval_int(vmg_ (long)(G_bif_tads_globals->rand_seed & 1));
}
#endif /* VMBIFTADS_RNG_BITSHIFT */


/* ------------------------------------------------------------------------ */
/*
 *   cvtstr (toString) - convert to string 
 */
void CVmBifTADS::cvtstr(VMG_ uint argc)
{
    const char *p;
    char buf[50];
    vm_val_t val;
    int radix;
    vm_val_t new_str;
    
    /* check arguments */
    check_argc_range(vmg_ argc, 1, 2);

    /* pop the argument */
    G_stk->pop(&val);

    /* if there's a radix specified, pop it as well */
    if (argc == 2)
    {
        /* get the radix from the stack */
        radix = pop_int_val(vmg0_);
    }
    else
    {
        /* use decimal by default */
        radix = 10;
    }

    /* convert the value */
    p = CVmObjString::cvt_to_str(vmg_ &new_str,
                                 buf, sizeof(buf), &val, radix);

    /* save the new string on the stack to protect from garbage collection */
    G_stk->push(&new_str);

    /* create and return a string from our new value */
    retval_obj(vmg_ CVmObjString::create(vmg_ FALSE,
                                         p + VMB_LEN, vmb_get_len(p)));

    /* done with the new string */
    G_stk->discard();
}

/*
 *   cvtnum (toInteger) - convert to an integer
 */
void CVmBifTADS::cvtnum(VMG_ uint argc)
{
    const char *strp;
    size_t len;
    int radix;
    vm_val_t *valp;
        
    /* check arguments */
    check_argc_range(vmg_ argc, 1, 2);

    /* 
     *   check for a BigNumber and convert it (not very object-oriented,
     *   but this is a type-conversion routine, so special awareness of
     *   individual types isn't that weird) 
     */
    valp = G_stk->get(0);
    if (valp->typ == VM_OBJ
        && CVmObjBigNum::is_bignum_obj(vmg_ valp->val.obj))
    {
        long intval;
        
        /* convert it as a BigNumber */
        intval = ((CVmObjBigNum *)vm_objp(vmg_ valp->val.obj))
                 ->convert_to_int();

        /* discard arguments (ignore the radix in this case) */
        G_stk->discard(argc);

        /* return the integer value */
        retval_int(vmg_ intval);
        return;
    }

    /* we can only convert a string */
    strp = pop_str_val(vmg0_);
    len = vmb_get_len(strp);

    /* if there's a radix specified, pop it as well */
    if (argc == 2)
    {
        /* get the radix from the stack */
        radix = pop_int_val(vmg0_);

        /* make sure the radix is valid */
        switch(radix)
        {
        case 8:
        case 10:
        case 16:
            /* it's okay - proceed */
            break;

        default:
            /* other radix values are invalid */
            err_throw(VMERR_BAD_VAL_BIF);
        }
    }
    else
    {
        /* the default radix is decimal */
        radix = 10;
    }

    /* parse the value */
    if (len == 3 && memcmp(strp + VMB_LEN, "nil", 3) == 0)
    {
        /* the value is the constant 'nil' */
        retval_nil(vmg0_);
    }
    else if (len == 4 && memcmp(strp + VMB_LEN, "true", 3) == 0)
    {
        /* the value is the constant 'true' */
        retval_true(vmg0_);
    }
    else
    {
        utf8_ptr p;
        size_t rem;
        int is_neg;
        ulong acc;

        /* scan past leading spaces */
        for (p.set((char *)strp + VMB_LEN), rem = len ;
             rem != 0 && is_space(p.getch()) ; p.inc(&rem)) ;

        /* presume it's positive */
        is_neg = FALSE;

        /* if the radix is 10, check for a leading + or - */
        if (radix == 10 && rem != 0)
        {
            if (p.getch() == '-')
            {
                /* note the sign and skip the character */
                is_neg = TRUE;
                p.inc(&rem);
            }
            else if (p.getch() == '+')
            {
                /* skip the character */
                p.inc(&rem);
            }
        }

        /* clear the accumulator */
        acc = 0;

        /* scan the digits */
        switch (radix)
        {
        case 8:
            for ( ; rem != 0 && is_odigit(p.getch()) ; p.inc(&rem))
            {
                acc <<= 3;
                acc += value_of_odigit(p.getch());
            }
            break;

        case 10:
            for ( ; rem != 0 && is_digit(p.getch()) ; p.inc(&rem))
            {
                acc *= 10;
                acc += value_of_digit(p.getch());
            }
            break;

        case 16:
            for ( ; rem != 0 && is_xdigit(p.getch()) ; p.inc(&rem))
            {
                acc <<= 4;
                acc += value_of_xdigit(p.getch());
            }
            break;
        }

        /* apply the sign, if appropriate, and set the return value */
        if (is_neg)
            retval_int(vmg_ -(long)acc);
        else
            retval_int(vmg_ (long)acc);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   put an integer value in a constant list, advancing the list write
 *   pointer 
 */
static void put_list_int(char **dstp, long intval)
{
    vm_val_t val;

    /* set up the integer value */
    val.set_int(intval);

    /* write it to the list */
    vmb_put_dh(*dstp, &val);

    /* advance the output pointer */
    *dstp += VMB_DATAHOLDER;
}

/*
 *   put an object value in a constant list, advancing the list write
 *   pointer 
 */
static void put_list_obj(char **dstp, vm_obj_id_t objval)
{
    vm_val_t val;

    /* set up the integer value */
    val.set_obj(objval);

    /* write it to the list */
    vmb_put_dh(*dstp, &val);

    /* advance the output pointer */
    *dstp += VMB_DATAHOLDER;
}


/*
 *   get the current time 
 */
void CVmBifTADS::gettime(VMG_ uint argc)
{
    int typ;
    time_t timer;
    struct tm *tblock;
    char buf[80];
    char *dst;
    
    /* check arguments */
    check_argc_range(vmg_ argc, 0, 1);

    /* if there's an argument, get the type of time value to return */
    if (argc == 1)
    {
        /* get the time type code */
        typ = pop_int_val(vmg0_);
    }
    else
    {
        /* use the default type */
        typ = 1;
    }

    /* check the type */
    switch(typ)
    {
    case 1:
        /* 
         *   default information - return the current time and date 
         */

        /* make sure the time zone is set up properly */
        os_tzset();

        /* get the local time information */
        timer = time(NULL);
        tblock = localtime(&timer);

        /* adjust values for return format */
        tblock->tm_year += 1900;
        tblock->tm_mon++;
        tblock->tm_wday++;
        tblock->tm_yday++;

        /*   
         *   build the return list: [year, month, day, day-of-week,
         *   day-of-year, hour, minute, second, seconds-since-1970] 
         */
        vmb_put_len(buf, 9);
        dst = buf + VMB_LEN;

        /* build return list value */
        put_list_int(&dst, tblock->tm_year);
        put_list_int(&dst, tblock->tm_mon);
        put_list_int(&dst, tblock->tm_mday);
        put_list_int(&dst, tblock->tm_wday);
        put_list_int(&dst, tblock->tm_yday);
        put_list_int(&dst, tblock->tm_hour);
        put_list_int(&dst, tblock->tm_min);
        put_list_int(&dst, tblock->tm_sec);
        put_list_int(&dst, (long)timer);

        /* allocate and return the list value */
        retval_obj(vmg_ CVmObjList::create(vmg_ FALSE, buf));

        /* done */
        break;

    case 2:
        /* high-precision system timer value */
        retval_int(vmg_ os_get_sys_clock_ms());
        break;

    default:
        err_throw(VMERR_BAD_VAL_BIF);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   re_match - match a regular expression to a string
 */
void CVmBifTADS::re_match(VMG_ uint argc)
{
    const char *pat;
    const char *str;
    int match_len;
    
    /* check arguments */
    check_argc(vmg_ argc, 2);

    /* get the pattern string */
    pat = pop_str_val(vmg0_);

    /* get the string to match to the pattern */
    str = pop_str_val(vmg0_);

    /* try matching the pattern */
    match_len = G_bif_tads_globals->regex->
                compile_and_match(pat + VMB_LEN, vmb_get_len(pat),
                                  str + VMB_LEN, vmb_get_len(str));

    /* check for a match */
    if (match_len >= 0)
    {
        utf8_ptr p;
        
        /* we got a match - calculate the character length of the match */
        p.set((char *)str + VMB_LEN);
        retval_int(vmg_ (long)p.len(match_len));
    }
    else
    {
        /* no match - return nil */
        retval_nil(vmg0_);
    }
}

/*
 *   re_search - search for a substring matching a regular expression
 *   within a string 
 */
void CVmBifTADS::re_search(VMG_ uint argc)
{
    vm_val_t strval;
    const char *pat;
    const char *str;
    int match_idx;
    int match_len;

    /* check arguments */
    check_argc(vmg_ argc, 2);

    /* get the pattern string */
    strval = *G_stk->get(0);
    pat = pop_str_val(vmg0_);

    /* get the string to match to the pattern */
    str = pop_str_val(vmg0_);

    /* push the string for GC protection */
    G_stk->push(&strval);

    /* try matching the pattern */
    match_idx = G_bif_tads_globals->regex->
                compile_and_search(pat + VMB_LEN, vmb_get_len(pat),
                                   str + VMB_LEN, vmb_get_len(str),
                                   &match_len);

    /* check for a match */
    if (match_idx >= 0)
    {
        utf8_ptr p;
        size_t char_idx;
        size_t char_len;
        vm_obj_id_t match_str_obj;
        char *dst;
        char buf[VMB_LEN + VMB_DATAHOLDER * 3];
        vm_val_t tmp_val;

        /* 
         *   We got a match - calculate the character index of the match
         *   offset, adjusted to a 1-base.  The character index is simply
         *   the number of characters in the part of the string up to the
         *   match index.  
         */
        p.set((char *)str + VMB_LEN);
        char_idx = p.len(match_idx) + 1;

        /* calculate the character length of the match */
        p.set((char *)str + VMB_LEN + match_idx);
        char_len = p.len(match_len);

        /* allocate a string containing the match */
        match_str_obj =
            CVmObjString::create(vmg_ FALSE, str + VMB_LEN + match_idx,
                                 match_len);

        /* push it momentarily as protection against garbage collection */
        tmp_val.set_obj(match_str_obj);
        G_stk->push(&tmp_val);

        /* 
         *   set up a 3-element list to contain the return value:
         *   [match_start_index, match_length, match_string] 
         */
        vmb_put_len(buf, 3);
        dst = buf + VMB_LEN;
        put_list_int(&dst, (long)char_idx);
        put_list_int(&dst, (long)char_len);
        put_list_obj(&dst, match_str_obj);

        /* allocate and return the list */
        retval_obj(vmg_ CVmObjList::create(vmg_ FALSE, buf));

        /* we no longer need the garbage collection protection */
        G_stk->discard();
    }
    else
    {
        /* no match - return nil */
        retval_nil(vmg0_);
    }

    /* discard the GC protection for the input string */
    G_stk->discard();
}

/*
 *   re_group - get the string matching a group in the most recent regular
 *   expression search or match 
 */
void CVmBifTADS::re_group(VMG_ uint argc)
{
    int groupno;
    const re_group_register *reg;
    char buf[VMB_LEN + 3*VMB_DATAHOLDER];
    char *dst;
    utf8_ptr p;
    vm_val_t tmp_val;
    
    /* check arguments */
    check_argc(vmg_ argc, 1);

    /* get the group number to retrieve */
    groupno = pop_int_val(vmg0_);

    /* make sure it's in range */
    if (groupno < 1 || groupno > RE_GROUP_REG_CNT)
        err_throw(VMERR_BAD_VAL_BIF);

    /* adjust from a 1-base to a 0-base */
    --groupno;

    /* if the group doesn't exist in the pattern, return nil */
    if (groupno >= G_bif_tads_globals->regex->get_group_cnt())
    {
        retval_nil(vmg0_);
        return;
    }

    /* get the register */
    reg = G_bif_tads_globals->regex->get_group_reg(groupno);

    /* if the group wasn't set, return nil */
    if (reg->start_ptr == 0 || reg->end_ptr == 0)
    {
        retval_nil(vmg0_);
        return;
    }
    
    /* set up for a list with three elements */
    vmb_put_len(buf, 3);
    dst = buf + VMB_LEN;

    /* 
     *   The first element is the character index of the group text in the
     *   source string.  Calculate the character index by adding 1 to the
     *   character length of the text preceding the group; calculate the
     *   character length from the byte length of that string.  
     */
    p.set((char *)G_bif_tads_globals->regex->get_last_str());
    put_list_int(&dst, p.len(reg->start_ptr - p.getptr()) + 1);

    /* 
     *   The second element is the character length of the group text.
     *   Calculate the character length from the byte length. 
     */
    p.set((char *)reg->start_ptr);
    put_list_int(&dst, p.len(reg->end_ptr - reg->start_ptr));

    /*
     *   The third element is the string itself.  Create a new string
     *   containing the matching substring. 
     */
    tmp_val.set_obj(CVmObjString::create(vmg_ FALSE, reg->start_ptr,
                                         reg->end_ptr - reg->start_ptr));
    put_list_obj(&dst, tmp_val.val.obj);

    /* save the string on the stack momentarily to protect against GC */
    G_stk->push(&tmp_val);

    /* create and return the list value */
    retval_obj(vmg_ CVmObjList::create(vmg_ FALSE, buf));

    /* we no longer need the garbage collector protection */
    G_stk->discard();
}

/*
 *   re_replace flags 
 */
#define VMBIFTADS_REPLACE_ALL  0x0001

/*
 *   re_replace - search for a pattern in a string, and apply a
 *   replacement pattern
 */
void CVmBifTADS::re_replace(VMG_ uint argc)
{
    vm_val_t patval, rplval;
    const char *pat;
    const char *str;
    const char *rpl;
    ulong flags;
    vm_val_t search_val;
    int match_idx;
    int match_len;
    size_t new_len;
    utf8_ptr p;
    size_t rem;
    int groupno;
    const re_group_register *reg;
    vm_obj_id_t ret_obj;
    utf8_ptr dstp;
    int match_cnt;
    int start_idx;

    /* check arguments */
    check_argc(vmg_ argc, 4);

    /* remember the pattern and replacement string values */
    patval = *G_stk->get(0);
    rplval = *G_stk->get(2);

    /* 
     *   Pop the pattern string, the search string, and the replacement
     *   string.  Note that we want to retain the original value
     *   information for the search string, since we'll end up returning
     *   it unchanged if we don't find the pattern.  
     */
    pat = pop_str_val(vmg0_);
    G_stk->pop(&search_val);
    rpl = pop_str_val(vmg0_);

    /* pop the flags */
    flags = pop_long_val(vmg0_);

    /* 
     *   put the pattern, replacement, and search string values back on
     *   the stack as protection against garbage collection 
     */
    G_stk->push(&patval);
    G_stk->push(&rplval);
    G_stk->push(&search_val);

    /* make sure the search string is indeed a string */
    str = search_val.get_as_string(vmg0_);
    if (str == 0)
        err_throw(VMERR_STRING_VAL_REQD);

    /*
     *   First, determine how long the result string will be.  Search
     *   repeatedly if the REPLACE_ALL flag (0x0001) is set.
     */
    for (new_len = 0, match_cnt = 0, start_idx = 0 ;
         (size_t)start_idx < vmb_get_len(str) ; ++match_cnt)
    {
        /* search for the pattern in the search string */
        match_idx = G_bif_tads_globals->regex->
                    compile_and_search(pat + VMB_LEN, vmb_get_len(pat),
                                       str + VMB_LEN + start_idx,
                                       vmb_get_len(str) - start_idx,
                                       &match_len);

        /* if there was no match, there is no more replacing to do */
        if (match_idx == -1)
        {
            /* 
             *   if we haven't found a match before, there's no
             *   replacement at all to do -- just return the original
             *   string unchanged 
             */
            if (match_cnt == 0)
            {
                /* no replacement - return the original search string */
                retval(vmg_ &search_val);
                goto done;
            }
            else
            {
                /* we've found all of our matches - stop searching */
                break;
            }
        }

        /*
         *   We've found a match to replace.  Determine how much space we
         *   need for the replacement pattern with its substitution
         *   parameters replaced with the original string's matching text.
         *   
         *   First, add in the length of the part from the start of this
         *   segment of the search to the matched substring.  
         */
        new_len += match_idx;

        /* 
         *   now, scan the replacement string and add in its length and
         *   the lengths of substitution parameters 
         */
        for (p.set((char *)rpl + VMB_LEN), rem = vmb_get_len(rpl) ;
             rem != 0 ; p.inc(&rem))
        {
            /* check for '%' sequences */
            if (p.getch() == '%')
            {
                /* skip the '%' */
                p.inc(&rem);
                
                /* if there's anything left, see what we have */
                if (rem != 0)
                {
                    switch(p.getch())
                    {
                    case '1':
                    case '2':
                    case '3':
                    case '4':
                    case '5':
                    case '6':
                    case '7':
                    case '8':
                    case '9':
                        /* get the group number */
                        groupno = value_of_digit(p.getch()) - 1;
                        
                        /* if this group is valid, add its length */
                        if (groupno
                            < G_bif_tads_globals->regex->get_group_cnt())
                        {
                            /* get the register */
                            reg = G_bif_tads_globals->regex
                                  ->get_group_reg(groupno);
                            
                            /* if it's been set, add its length */
                            if (reg->start_ptr != 0 && reg->end_ptr != 0)
                                new_len += reg->end_ptr - reg->start_ptr;
                        }
                        break;
                        
                    case '*':
                        /* add the entire match size */
                        new_len += match_len;
                        break;
                        
                    case '%':
                        /* add a single '%' */
                        ++new_len;
                        break;
                        
                    default:
                        /* add the entire sequence unchanged */
                        new_len += 2;
                        break;
                    }
                }
            }
            else
            {
                /* count this character literally */
                new_len += p.charsize();
            }
        }

        /* start the next search after the end of this match */
        start_idx += match_idx + match_len;

        /* 
         *   if we're only replacing a single match, stop now; otherwise,
         *   continue looking 
         */
        if (!(flags & VMBIFTADS_REPLACE_ALL))
            break;
    }

    /* add in the size of the remainder of the string after the last match */
    new_len += vmb_get_len(str) - start_idx;

    /* allocate the result string */
    ret_obj = CVmObjString::create(vmg_ FALSE, new_len);

    /* get a pointer to the result buffer */
    dstp.set(((CVmObjString *)vm_objp(vmg_ ret_obj))->cons_get_buf());

    /*
     *   Once again, start searching from the beginning of the string.
     *   This time, build the result string as we go. 
     */
    for (start_idx = 0 ; (size_t)start_idx < vmb_get_len(str) ; )
    {
        /* search for the pattern */
        match_idx = G_bif_tads_globals->regex->
                    compile_and_search(pat + VMB_LEN, vmb_get_len(pat),
                                       str + VMB_LEN + start_idx,
                                       vmb_get_len(str) - start_idx,
                                       &match_len);

        /* stop if we can't find another match */
        if (match_idx < 0)
            break;
        
        /* copy the part up to the start of the matched text, if any */
        if (match_idx > 0)
        {
            /* copy the part from the last match to this match */
            memcpy(dstp.getptr(), str + VMB_LEN + start_idx, match_idx);

            /* advance the output pointer */
            dstp.set(dstp.getptr() + match_idx);
        }

        /*
         *   Scan the replacement string again, and this time actually
         *   build the result.  
         */
        for (p.set((char *)rpl + VMB_LEN), rem = vmb_get_len(rpl) ;
             rem != 0 ; p.inc(&rem))
        {
            /* check for '%' sequences */
            if (p.getch() == '%')
            {
                /* skip the '%' */
                p.inc(&rem);
                
                /* if there's anything left, see what we have */
                if (rem != 0)
                {
                    switch(p.getch())
                    {
                    case '1':
                    case '2':
                    case '3':
                    case '4':
                    case '5':
                    case '6':
                    case '7':
                    case '8':
                    case '9':
                        /* get the group number */
                        groupno = value_of_digit(p.getch()) - 1;
                        
                        /* if this group is valid, add its length */
                        if (groupno
                            < G_bif_tads_globals->regex->get_group_cnt())
                        {
                            /* get the register */
                            reg = G_bif_tads_globals->regex
                                  ->get_group_reg(groupno);
                            
                            /* if it's been set, add its text */
                            if (reg->start_ptr != 0 && reg->end_ptr != 0)
                            {
                                /* copy the data */
                                memcpy(dstp.getptr(), reg->start_ptr,
                                       reg->end_ptr - reg->start_ptr);

                                /* advance past it */
                                dstp.set(dstp.getptr()
                                         + (reg->end_ptr - reg->start_ptr));
                            }
                        }
                        break;

                    case '*':
                        /* add the entire matched string */
                        memcpy(dstp.getptr(),
                               str + VMB_LEN + start_idx + match_idx,
                               match_len);
                        dstp.set(dstp.getptr() + match_len);
                        break;
                        
                    case '%':
                        /* add a single '%' */
                        dstp.setch('%');
                        break;
                        
                    default:
                        /* add the entire sequence unchanged */
                        dstp.setch('%');
                        dstp.setch(p.getch());
                        break;
                    }
                }
            }
            else
            {
                /* copy this character literally */
                dstp.setch(p.getch());
            }
        }

        /* advance past this matched string for the next search */
        start_idx += match_idx + match_len;

        /* if we're only performing a single replacement, stop now */
        if (!(flags & VMBIFTADS_REPLACE_ALL))
            break;
    }

    /* add the part after the end of the matched text */
    if ((size_t)start_idx < vmb_get_len(str))
        memcpy(dstp.getptr(), str + VMB_LEN + start_idx,
               vmb_get_len(str) - start_idx);

    /* return the string */
    retval_obj(vmg_ ret_obj);

done:
    /* discard the garbage collection protection references */
    G_stk->discard(3);
}

/* ------------------------------------------------------------------------ */
/*
 *   savepoint - establish an undo savepoint
 */
void CVmBifTADS::savepoint(VMG_ uint argc)
{
    /* check arguments */
    check_argc(vmg_ argc, 0);

    /* establish the savepoint */
    G_undo->create_savept(vmg0_);
}

/*
 *   undo - undo changes to most recent savepoint
 */
void CVmBifTADS::undo(VMG_ uint argc)
{
    /* check arguments */
    check_argc(vmg_ argc, 0);

    /* if no undo is available, return nil to indicate that we can't undo */
    if (G_undo->get_savept_cnt() == 0)
    {
        /* we can't undo */
        retval_nil(vmg0_);
    }
    else
    {
        /* undo to the savepoint */
        G_undo->undo_to_savept(vmg0_);

        /* tell the caller that we succeeded */
        retval_true(vmg0_);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   save 
 */
void CVmBifTADS::save(VMG_ uint argc)
{
    char fname[OSFNMAX];
    CVmFile *file;
    osfildef *fp;
    
    /* check arguments */
    check_argc(vmg_ argc, 1);

    /* get the filename as a null-terminated string */
    pop_str_val_fname(vmg_ fname, sizeof(fname));

    /* open the file */
    fp = osfopwb(fname, OSFTT3SAV);
    if (fp == 0)
        err_throw(VMERR_CREATE_FILE);

    /* set up the file writer */
    file = new CVmFile();
    file->set_file(fp, 0);

    err_try
    {
        /* save the state */
        CVmSaveFile::save(vmg_ file);
    }
    err_finally
    {
        /* close the file */
        delete file;
    }
    err_end;

    /* 
     *   return nil to indicate that we have saved the state (this is a
     *   setjmp-style indication that we are still executing the program
     *   after the save, as opposed to having just restored the program,
     *   at which point we'll also look like we just returned from the
     *   'save' call) 
     */
    retval_nil(vmg0_);
}

/*
 *   restore
 */
void CVmBifTADS::restore(VMG_ uint argc)
{
    char fname[OSFNMAX];
    CVmFile *file;
    osfildef *fp;
    vm_val_t rval;
    int err;

    /* check arguments */
    check_argc(vmg_ argc, 2);

    /* get the filename as a null-terminated string */
    pop_str_val_fname(vmg_ fname, sizeof(fname));

    /* 
     *   Get the return value - when we're done, we'll return this from
     *   what will look like the call to 'save', setjmp-style, to tell the
     *   'save' caller that we actually just restored the state.  This
     *   value must be an integer.  
     */
    G_interpreter->pop_int(vmg_ &rval);

    /* open the file */
    fp = osfoprb(fname, OSFTT3SAV);
    if (fp == 0)
        err_throw(VMERR_FILE_NOT_FOUND);

    /* set up the file reader */
    file = new CVmFile();
    file->set_file(fp, 0);

    err_try
    {
        /* restore the state */
        err = CVmSaveFile::restore(vmg_ file);
    }
    err_finally
    {
        /* close the file */
        delete file;
    }
    err_end;

    /* if an error occurred, throw an exception */
    if (err != 0)
        err_throw(err);

    /* 
     *   Return the value that the caller asked us to return.  Now that
     *   we've restored the original state, it will look as though we're
     *   returning from the 'save()' call that created this saved state
     *   file, because things are now back to exactly as they were at that
     *   point.  The caller will be able to tell that we just restored the
     *   state, though, by looking at the return value.  
     */
    retval(vmg_ &rval);
}

/*
 *   restart
 */
void CVmBifTADS::restart(VMG_ uint argc)
{
    vm_val_t func_ptr;
    vm_val_t func_arg;
    
    /* check arguments */
    check_argc(vmg_ argc, 2);

    /* 
     *   pop the function pointer (this is the function we call after
     *   restarting) and the integer argument to the function 
     */
    G_interpreter->pop_funcptr(vmg_ &func_ptr);
    G_interpreter->pop_int(vmg_ &func_arg);

    /* reset the VM to the image file's initial state */
    CVmSaveFile::reset(vmg0_);

    /* push the argument to the re-initialization function */
    G_stk->push(&func_arg);

    /* set up a call frame for the function we're calling */
    G_interpreter->set_up_recursive_call_frame(vmg_ func_ptr.val.ofs, 1,
                                               VM_INVALID_OBJ);
}


/* ------------------------------------------------------------------------ */
/*
 *   Get the maximum value from a set of argument 
 */
void CVmBifTADS::get_max(VMG_ uint argc)
{
    uint i;
    vm_val_t cur_max;
    
    /* make sure we have at least one argument */
    if (argc < 1)
        err_throw(VMERR_WRONG_NUM_OF_ARGS);

    /* start with the first argument as the presumptive maximum */
    cur_max = *G_stk->get(0);

    /* compare each argument in turn */
    for (i = 1 ; i < argc ; ++i)
    {
        /* 
         *   compare this value to the maximum so far; if this value is
         *   greater, it becomes the new maximum so far 
         */
        if (G_stk->get(i)->compare_to(vmg_ &cur_max) > 0)
            cur_max = *G_stk->get(i);
    }

    /* discard the arguments */
    G_stk->discard(argc);

    /* return the maximum value */
    retval(vmg_ &cur_max);
}

/*
 *   Get the minimum value from a set of argument 
 */
void CVmBifTADS::get_min(VMG_ uint argc)
{
    uint i;
    vm_val_t cur_min;

    /* make sure we have at least one argument */
    if (argc < 1)
        err_throw(VMERR_WRONG_NUM_OF_ARGS);

    /* start with the first argument as the presumptive minimum */
    cur_min = *G_stk->get(0);

    /* compare each argument in turn */
    for (i = 1 ; i < argc ; ++i)
    {
        /* 
         *   compare this value to the minimum so far; if this value is
         *   less, it becomes the new minimum so far 
         */
        if (G_stk->get(i)->compare_to(vmg_ &cur_min) < 0)
            cur_min = *G_stk->get(i);
    }

    /* discard the arguments */
    G_stk->discard(argc);

    /* return the minimum value */
    retval(vmg_ &cur_min);
}

/* ------------------------------------------------------------------------ */
/*
 *   makeString - construct a string by repeating a character; by
 *   converting a unicode code point to a string; or by converting a list
 *   of unicode code points to a string 
 */
void CVmBifTADS::make_string(VMG_ uint argc)
{
    vm_val_t val;
    long rpt;
    vm_obj_id_t new_str_obj;
    CVmObjString *new_str;
    size_t new_str_len;
    char *new_strp;
    const char *lstp = 0;
    const char *strp = 0;
    size_t len;
    size_t i;
    utf8_ptr dst;
    
    /* check arguments */
    check_argc_range(vmg_ argc, 1, 2);

    /* get the base value */
    G_stk->pop(&val);

    /* if there's a repeat count, get it */
    rpt = (argc >= 2 ? pop_long_val(vmg0_) : 1);

    /* if the repeat count is less than or equal to zero, make it 1 */
    if (rpt < 1)
        rpt = 1;

    /* leave the original value on the stack to protect it from GC */
    G_stk->push(&val);

    /* 
     *   see what we have, and calculate how much space we'll need for the
     *   result string 
     */
    switch(val.typ)
    {
    case VM_LIST:
        /* it's a list of integers giving unicode character values */
        lstp = G_const_pool->get_ptr(val.val.ofs);

    do_list:
        /* get the list count */
        len = vmb_get_len(lstp);

        /* 
         *   Run through the list and get the size of each character, so
         *   we can determine how long the string will have to be. 
         */
        for (new_str_len = 0, i = 1 ; i <= len ; ++i)
        {
            vm_val_t ele_val;
            
            /* get this element */
            CVmObjList::index_list(vmg_ &ele_val, lstp, i);

            /* if it's not an integer, it's an error */
            if (ele_val.typ != VM_INT)
                err_throw(VMERR_INT_VAL_REQD);

            /* add this character's byte size to the string size */
            new_str_len +=
                utf8_ptr::s_wchar_size((wchar_t)ele_val.val.intval);
        }
        break;

    case VM_SSTRING:
        /* get the string pointer */
        strp = G_const_pool->get_ptr(val.val.ofs);
        
    do_string:
        /* 
         *   it's a string - the output length is the same as the input
         *   length 
         */
        new_str_len = vmb_get_len(strp);
        break;

    case VM_INT:
        /* 
         *   it's an integer giving a unicode character value - we just
         *   need enough space to store this particular character 
         */
        new_str_len = utf8_ptr::s_wchar_size((wchar_t)val.val.intval);
        break;

    case VM_OBJ:
        /* check to see if it's a string */
        if ((strp = val.get_as_string(vmg0_)) != 0)
            goto do_string;

        /* check to see if it's a list */
        if ((lstp = val.get_as_list(vmg0_)) != 0)
            goto do_list;

        /* it's invalid */
        err_throw(VMERR_BAD_TYPE_BIF);
        break;

    default:
        /* other types are invalid */
        err_throw(VMERR_BAD_TYPE_BIF);
        break;
    }

    /* 
     *   if the length times the repeat count would be over the maximum
     *   16-bit string length, it's an error 
     */
    if (new_str_len * rpt > 0xffffL - VMB_LEN)
        err_throw(VMERR_BAD_VAL_BIF);

    /* multiply the length by the repeat count */
    new_str_len *= rpt;
    
    /* allocate the string and gets its buffer */
    new_str_obj = CVmObjString::create(vmg_ FALSE, new_str_len);
    new_str = (CVmObjString *)vm_objp(vmg_ new_str_obj);
    new_strp = new_str->cons_get_buf();
    
    /* set up the destination pointer */
    dst.set(new_strp);

    /* run through the number of iterations requested */
    for ( ; rpt != 0 ; --rpt)
    {
        /* build one iteration of the string, according to the type */
        if (lstp != 0)
        {
            /* run through the list */
            for (i = 1 ; i <= len ; ++i)
            {
                vm_val_t ele_val;
                
                /* get this element */
                CVmObjList::index_list(vmg_ &ele_val, lstp, i);

                /* add this character to the string */
                dst.setch((wchar_t)ele_val.val.intval);
            }
        }
        else if (strp != 0)
        {
            /* copy the string's contents into the output string */
            memcpy(dst.getptr(), strp + VMB_LEN, vmb_get_len(strp));

            /* advance past the bytes we copied */
            dst.set(dst.getptr() + vmb_get_len(strp));
        }
        else
        {
            /* set this int value */
            dst.setch((wchar_t)val.val.intval);
        }
    }

    /* return the new string */
    retval_obj(vmg_ new_str_obj);

    /* discard the GC protection */
    G_stk->discard();
}

